==> lidR_3.2.0.tar.gz: lidR/R/projection.R <==
# ===============================================================================
#
# PROGRAMMERS:
#
# jean-romain.roussel.1@ulaval.ca  -  https://github.com/Jean-Romain/lidR
#
# COPYRIGHT:
#
# Copyright 2016-2019 Jean-Romain Roussel
#
# This file is part of lidR R package.
#
# lidR is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#
# ===============================================================================

#' Get or set the projection of a LAS* object
#'
#' Get or set the projection of a LAS* object.
#'
#' There are two ways to store the CRS of a point cloud in a LAS file:
#'
#'    - Store an EPSG code (for LAS 1.0 to 1.3)
#'    - Store a WTK string (for LAS 1.4)
#'
#' On the other hand, R sptial packages use a `proj4string` to store the CRS (but
#' the ecosystem is moving to WKT). This is why the CRS is duplicated in a LAS object.
#' The information belongs within the header in a format that can be written in a
#' LAS file and in the slot `proj4string` in a format that can be understood by R
#' packages.
#'
#'    - `projection<-`: assigns a CRS from a `CRS` (`sp`), a `crs` (`sf`), a WKT
#'    string, a proj4string or an epsg code. It updates the header of the LAS
#'    object either with the EPSG code for LAS formats < 1.4 or with a WKT string
#'    for LAS format 1.4 and updates the `proj4string` slot.
#'    - `projection`: returns the CRS in `sp` format
#'    - `crs` and `crs<-` are equivalent to `projection` and `projection<-`
#'    - `epsg<-`, `wkt<- `: legacy functions superseded by `projection<-`
#'    - `sf::st_crs` return the CRS in `sf` format.
#'    - `epsg`: reads the epsg code from the header.
#'    - `wkt`: reads the WKT string from the header.
#'
#' @param object,x,obj An object of class LAS or eventually LASheader (regular users don't need to manipulate
#' LASheader objects).
#' @param ... Unused.
#' @param asText logical. If TRUE, the projection is returned as text. Otherwise a CRS object is returned.
#' @param value A `CRS` object or a `proj4string` string or WKT string or an EPSG code.
#'
#' @export
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile)
#'
#' # Get the EPSG code stored in the header (returns 0 if not recorded)
#' epsg(las)
#'
#' # Get the WKT string stored in the header (LAS >= 1.4)
#' wkt(las)
#'
#' # Overwrite the CRS (but does not reproject)
#' crs <- sp::CRS("+init=epsg:26918")
#' projection(las) <- crs
#' sf::st_crs(las)$input
#'
#' # Uses the EPSG code
#' projection(las) <- 26919
#' sf::st_crs(las)$input
#'
#' # Uses a crs from sf
#' crs <- sf::st_crs(3035)
#' projection(las) <- crs
#'
#' @importFrom raster projection<-
#' @importFrom raster projection
#' @importFrom raster crs<-
#' @importFrom raster crs
#' @name projection
#' @rdname projection
#' @md
NULL

# ==== EPSG ====

#' @export
#' @rdname projection
setGeneric("epsg", function(object, ...)
  standardGeneric("epsg"))

#' @export
#' @rdname projection
setGeneric("epsg<-", function(object, value)
  standardGeneric("epsg<-"))

#' @export
#' @rdname projection
setMethod("epsg", "LASheader", function(object, ...)
{
  if (use_wktcs(object))  warning("This LAS object stores the CRS as WKT. 0 returned; use 'wkt()' instead.", call. = FALSE)
  return(rlas::header_get_epsg(as.list(object)))
})

#' @export
#' @rdname projection
setMethod("epsg<-", "LASheader", function(object, value)
{
  if (use_wktcs(object)) stop("This object is not in a format that supports EPSG code. Use a WKT string.", call. = FALSE)
  header <- as.list(object)
  header <- rlas::header_set_epsg(header, value)
  return(LASheader(header))
})

#' @export
#' @rdname projection
setMethod("epsg", "LAS", function(object)
{
  return(epsg(object@header))
})

#' @export
#' @rdname projection
setMethod("epsg<-", "LAS", function(object, value)
{
  proj4 <- epsg2CRS(value, fail = TRUE)
  epsg(object@header) <- value
  object@proj4string <- proj4
  return(object)
})

# ==== WKT =====

# @export
# @rdname projection
#setGeneric("wkt", function(object, ...)
#  standardGeneric("wkt"))

#' @export
#' @rdname projection
setGeneric("wkt<-", function(object, value)
  standardGeneric("wkt<-"))

#' @export
#' @importFrom sp wkt
#' @rdname projection
setMethod("wkt", "LASheader", function(obj)
{
  return(rlas::header_get_wktcs(as.list(obj)))
})

#' @export
#' @rdname projection
setMethod("wkt<-", "LASheader", function(object, value)
{
  if (!use_wktcs(object)) stop("This object is not in a format that supports WKT CRS. Use an EPSG code.", call. = FALSE)
  header <- as.list(object)
  header <- rlas::header_set_wktcs(header, value)
  return(LASheader(header))
})

#' @export
#' @importFrom sp wkt
#' @rdname projection
setMethod("wkt", "LAS", function(obj)
{
  return(wkt(obj@header))
})

#' @export
#' @rdname projection
setMethod("wkt<-", "LAS", function(object, value)
{
  proj4 <- wkt2CRS(value, fail = TRUE)
  wkt(object@header) <- value
  object@proj4string <- proj4
  return(object)
})


# ==== PROJECTION =====

#' @export
#' @rdname projection
setMethod("projection", "LASheader", function(x, asText = TRUE)
{
  if (use_epsg(x) && epsg(x) != 0L)
    proj4 <- epsg2CRS(epsg(x))
  else if (use_wktcs(x) && wkt(x) != "")
  {
    proj4 <- wkt2CRS(wkt(x))
  }
  else
    proj4 <- sp::CRS()

  if (asText)
    return(proj4@projargs)
  else
    return(proj4)
})

# This is legacy code that should no longer be used because it returns a proj4 string

#' @export
#' @rdname projection
setMethod("projection", "LAS", function(x, asText = TRUE)
{
  proj4 <- x@proj4string

  if (asText)
    return(proj4@projargs)
  else
    return(proj4)
})

#' @export
#' @rdname projection
setMethod("projection<-", "LAS", function(x, value)
{
  # The input is either:
  # - a string > proj4 or WKT
  # - a CRS > from sp
  # - a crs > from sf
  # - a number > EPSG code
  # In all case we need to get a WKT or an EPSG code to update the
  # header depending on the LAS format (1.4 or above)
  # In addition we need this to work both with old and new proj and gdal

  all_crs <- all_crs_formats(value)
  CRS <- all_crs[["CRS"]]
  wkt <- all_crs[["wkt"]]
  epsg <- all_crs[["epsg"]]
  proj4 <- all_crs[["proj4"]]

  if (is.null(epsg) | is.na(epsg))
    epsg = 0

  if (use_wktcs(x))
  {
    if (is.null(wkt) || wkt == "")
      x@header@VLR[["WKT OGC CS"]] <- NULL
    else
      wkt(x@header) <- wkt

    x@proj4string <- CRS
  }
  else
  {
    if (epsg == 0L)
    {
      if(!is.na(proj4)) warning(paste0("EPSG code not found: header not updated. Try to provide the ESPG code instead of a ", class(value)[1]), call. = FALSE)
      pos <- rlas:::where_is_epsg(as.list(x@header))
      if (pos != 0)  x@header@VLR[["GeoKeyDirectoryTag"]][["tags"]][[pos]] <- NULL
    }
    else
    {
      epsg(x@header) <- epsg
    }

    x@proj4string <- CRS
  }

  return(x)
})

#' @export
#' @rdname projection
setMethod("projection", "LAScatalog", function(x, asText = TRUE)
{
  proj4 <- x@proj4string

  if (asText)
    return(proj4@projargs)
  else
    return(proj4)
})

#' @export
#' @rdname projection
setMethod("projection<-", "LAScatalog", function(x, value)
{
  all_crs <- all_crs_formats(value)
  x@proj4string <- all_crs[["CRS"]]
  return(x)
})


# ==== CRS ====

#' @export
#' @rdname projection
setMethod("crs", "LASheader", function(x, asText = FALSE)
{
  return(projection(x, asText))
})

#' @export
#' @rdname projection
setMethod("crs", "LAS", function(x, asText = FALSE)
{
  return(projection(x, asText))
})

#' @export
#' @rdname projection
setMethod("crs<-", "LAS", function(x, ..., value)
{
  projection(x) <- value
  return(x)
})


#' @export
#' @rdname projection
setMethod("crs", "LAScatalog", function(x, asText = FALSE)
{
  return(projection(x, asText))
})

#' @export
#' @rdname projection
setMethod("crs<-", "LAScatalog", function(x, ..., value)
{
  projection(x) <- value
  return(x)
})


# ===== SPTRANSFORM ====

#' @importMethodsFrom sp spTransform
#' @noRd
setMethod("spTransform", signature("LAS", "CRS"), function(x, CRSobj, ...)
{
  if (is.na(x@proj4string@projargs))
    stop("No transformation possible from NA reference system")

  p <- list(...)

  # Transform the point coordinates
  spts <- sp::SpatialPoints(coordinates(x))
  spts@proj4string <- crs(x)
  spts <- sp::spTransform(spts, CRSobj)
  X <- spts@coords[,1]
  Y <- spts@coords[,2]

  # Update the offsets in the header
  offsetx <- floor(min(X))
  offsety <- floor(min(Y))
  if (!is.null(p$xoffset)) offsetx <- p$xoffset
  if (!is.null(p$yoffset)) offsety <- p$yoffset
  x@header@PHB[["X offset"]] <- offsetx
  x@header@PHB[["Y offset"]] <- offsety

  # Update the scale factors
  scalex <- x@header@PHB[["X scale factor"]]
  scaley <- x@header@PHB[["Y scale factor"]]
  if (!is.null(p$scale)) scalex <- scaley <- p$scale
  x@header@PHB[["X scale factor"]] <- scalex
  x@header@PHB[["Y scale factor"]] <- scaley

  # Quantize the coordinates
  fast_quantization(X, scalex, offsetx)
  fast_quantization(Y, scaley, offsety)

  # Update the LAS object
  x@data[["X"]] <- X
  x@data[["Y"]] <- Y
  x <- lasupdateheader(x)
  crs(x) <- CRSobj

  return(x)
})

# ===== INTERNAL TOOLS =======

use_wktcs <- function(x) {
  UseMethod("use_wktcs", x)
}

use_wktcs.LAS <- function(x) {
  return(x@header@PHB[["Global Encoding"]][["WKT"]])
}

use_wktcs.LASheader <- function(x) {
  return(x@PHB[["Global Encoding"]][["WKT"]])
}

use_epsg <- function(x) {
  UseMethod("use_epsg", x)
}

use_epsg.LAS <- function(x) {
  return(!x@header@PHB[["Global Encoding"]][["WKT"]])
}

use_epsg.LASheader <- function(x) {
  return(!x@PHB[["Global Encoding"]][["WKT"]])
}

all_crs_formats = function(value)
{
  if (is(value, "CRS"))
  {
    proj4 <- value@projargs
    CRS <- value
    crs <- sf::st_crs(CRS)
    wkt <- crs$wkt
    epsg <- crs$epsg
  }
  else if (is(value, "crs"))
  {
    wkt <- value$wkt
    CRS <- sp::CRS(SRS_string = wkt)
    crs <- value
    proj4 <- CRS@projargs
    epsg <- crs$epsg
  }
  else if (is.character(value))
  {
    CRS <- sp::CRS(SRS_string = value)
    proj4 <- CRS@projargs
    crs <- sf::st_crs(CRS)
    wkt <- crs$wkt
    epsg <- crs$epsg
  }
  else if (is.numeric(value))
  {
    epsg <- value
    CRS <- epsg2CRS(epsg)
    crs <- sf::st_crs(CRS)
    proj4 <- CRS@projargs
    wkt <- crs$wkt
  }
  else
    stop("'value' is not a CRS or a string or a number.")

  return(list(epsg = epsg, CRS = CRS, crs = crs, proj4 = proj4, wkt = wkt))
}


epsg2CRS <- function(epsg, fail = FALSE)
{
  crs <- tryCatch(
  {
    crs <- suppressWarnings(sf::st_crs(paste0("EPSG:", epsg)))
    return(suppressWarnings(as(crs, "CRS")))
  },
  error = function(e)
  {
    if (!fail)
      return(sp::CRS(NA_character_))
    else
      stop(paste("Invalid epsg code", epsg), call. = FALSE)
  })

  return(crs)
}

wkt2CRS <- function(wkt, fail = FALSE)
{
  crs <- tryCatch(
  {
    as(sf::st_crs(wkt), "CRS")
  },
  error = function(e)
  {
    if (!fail)
      return(sp::CRS(NA_character_))
    else
      stop("Invalid WKT string", call. = FALSE)
  })

  return(crs)
}

==> lidR_3.2.2.tar.gz: lidR/R/projection.R <==
# ===============================================================================
#
# PROGRAMMERS:
#
# jean-romain.roussel.1@ulaval.ca  -  https://github.com/Jean-Romain/lidR
#
# COPYRIGHT:
#
# Copyright 2016-2019 Jean-Romain Roussel
#
# This file is part of lidR R package.
#
# lidR is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#
# ===============================================================================

#' Get or set the projection of a LAS* object
#'
#' Get or set the projection of a LAS* object.
#'
#' There are two ways to store the CRS of a point cloud in a LAS file:
#'
#'    - Store an EPSG code (for LAS 1.0 to 1.3)
#'    - Store a WTK string (for LAS 1.4)
#'
#' On the other hand, R sptial packages use a `proj4string` to store the CRS (but
#' the ecosystem is moving to WKT). This is why the CRS is duplicated in a LAS object.
#' The information belongs within the header in a format that can be written in a
#' LAS file and in the slot `proj4string` in a format that can be understood by R
#' packages.
#'
#'    - `projection<-`: assigns a CRS from a `CRS` (`sp`), a `crs` (`sf`), a WKT
#'    string, a proj4string or an epsg code. It updates the header of the LAS
#'    object either with the EPSG code for LAS formats < 1.4 or with a WKT string
#'    for LAS format 1.4 and updates the `proj4string` slot.
#'    - `projection`: returns the CRS in `sp` format
#'    - `crs` and `crs<-` are equivalent to `projection` and `projection<-`
#'    - `epsg<-`, `wkt<- `: legacy functions superseded by `projection<-`
#'    - `sf::st_crs` return the CRS in `sf` format.
#'    - `epsg`: reads the epsg code from the header.
#'    - `wkt`: reads the WKT string from the header.
#'
#' @param object,x,obj An object of class LAS or eventually LASheader (regular users don't need to manipulate
#' LASheader objects).
#' @param ... Unused.
#' @param asText logical. If TRUE, the projection is returned as text. Otherwise a CRS object is returned.
#' @param value A `CRS` object or a `proj4string` string or WKT string or an EPSG code.
#'
#' @export
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile)
#'
#' # Get the EPSG code stored in the header (returns 0 if not recorded)
#' epsg(las)
#'
#' # Get the WKT string stored in the header (LAS >= 1.4)
#' wkt(las)
#'
#' # Uses the EPSG code
#' projection(las) <- 26919
#' sf::st_crs(las)$input
#'
#' @importFrom raster projection<-
#' @importFrom raster projection
#' @importFrom raster crs<-
#' @importFrom raster crs
#' @name projection
#' @rdname projection
#' @md
NULL

# ==== EPSG ====

#' @export
#' @rdname projection
setGeneric("epsg", function(object, ...)
  standardGeneric("epsg"))

#' @export
#' @rdname projection
setGeneric("epsg<-", function(object, value)
  standardGeneric("epsg<-"))

#' @export
#' @rdname projection
setMethod("epsg", "LASheader", function(object, ...)
{
  if (use_wktcs(object))  warning("This LAS object stores the CRS as WKT. 0 returned; use 'wkt()' instead.", call. = FALSE)
  return(rlas::header_get_epsg(as.list(object)))
})

#' @export
#' @rdname projection
setMethod("epsg<-", "LASheader", function(object, value)
{
  if (use_wktcs(object)) stop("This object is not in a format that supports EPSG code. Use a WKT string.", call. = FALSE)
  header <- as.list(object)
  header <- rlas::header_set_epsg(header, value)
  return(LASheader(header))
})

#' @export
#' @rdname projection
setMethod("epsg", "LAS", function(object)
{
  return(epsg(object@header))
})

#' @export
#' @rdname projection
setMethod("epsg<-", "LAS", function(object, value)
{
  proj4 <- epsg2CRS(value, fail = TRUE)
  epsg(object@header) <- value
  object@proj4string <- proj4
  return(object)
})

# ==== WKT =====

# @export
# @rdname projection
#setGeneric("wkt", function(object, ...)
#  standardGeneric("wkt"))

#' @export
#' @rdname projection
setGeneric("wkt<-", function(object, value)
  standardGeneric("wkt<-"))

#' @export
#' @importFrom sp wkt
#' @rdname projection
setMethod("wkt", "LASheader", function(obj)
{
  return(rlas::header_get_wktcs(as.list(obj)))
})

#' @export
#' @rdname projection
setMethod("wkt<-", "LASheader", function(object, value)
{
  if (!use_wktcs(object)) stop("This object is not in a format that supports WKT CRS. Use an EPSG code.", call. = FALSE)
  header <- as.list(object)
  header <- rlas::header_set_wktcs(header, value)
  return(LASheader(header))
})

#' @export
#' @importFrom sp wkt
#' @rdname projection
setMethod("wkt", "LAS", function(obj)
{
  return(wkt(obj@header))
})

#' @export
#' @rdname projection
setMethod("wkt<-", "LAS", function(object, value)
{
  proj4 <- wkt2CRS(value, fail = TRUE)
  wkt(object@header) <- value
  object@proj4string <- proj4
  return(object)
})


# ==== PROJECTION =====

#' @export
#' @rdname projection
setMethod("projection", "LASheader", function(x, asText = TRUE)
{
  if (use_epsg(x) && epsg(x) != 0L)
    proj4 <- epsg2CRS(epsg(x))
  else if (use_wktcs(x) && wkt(x) != "")
  {
    proj4 <- wkt2CRS(wkt(x))
  }
  else
    proj4 <- sp::CRS()

  if (asText)
    return(proj4@projargs)
  else
    return(proj4)
})

# This is legacy code that should no longer be used because it returns a proj4 string

#' @export
#' @rdname projection
setMethod("projection", "LAS", function(x, asText = TRUE)
{
  proj4 <- x@proj4string

  if (asText)
    return(proj4@projargs)
  else
    return(proj4)
})

#' @export
#' @rdname projection
setMethod("projection<-", "LAS", function(x, value)
{
  # The input is either:
  # - a string > proj4 or WKT
  # - a CRS > from sp
  # - a crs > from sf
  # - a number > EPSG code
  # In all case we need to get a WKT or an EPSG code to update the
  # header depending on the LAS format (1.4 or above)
  # In addition we need this to work both with old and new proj and gdal

  all_crs <- all_crs_formats(value)
  CRS <- all_crs[["CRS"]]
  wkt <- all_crs[["wkt"]]
  epsg <- all_crs[["epsg"]]
  proj4 <- all_crs[["proj4"]]

  if (is.null(epsg) | is.na(epsg))
    epsg = 0

  if (use_wktcs(x))
  {
    if (is.null(wkt) || wkt == "")
      x@header@VLR[["WKT OGC CS"]] <- NULL
    else
      wkt(x@header) <- wkt

    x@proj4string <- CRS
  }
  else
  {
    if (epsg == 0L)
    {
      if(!is.na(proj4)) warning(paste0("EPSG code not found: header not updated. Try to provide the ESPG code instead of a ", class(value)[1]), call. = FALSE)
      pos <- rlas:::where_is_epsg(as.list(x@header))
      if (pos != 0)  x@header@VLR[["GeoKeyDirectoryTag"]][["tags"]][[pos]] <- NULL
    }
    else
    {
      epsg(x@header) <- epsg
    }

    x@proj4string <- CRS
  }

  return(x)
})

#' @export
#' @rdname projection
setMethod("projection", "LAScatalog", function(x, asText = TRUE)
{
  proj4 <- x@proj4string

  if (asText)
    return(proj4@projargs)
  else
    return(proj4)
})

#' @export
#' @rdname projection
setMethod("projection<-", "LAScatalog", function(x, value)
{
  all_crs <- all_crs_formats(value)
  x@proj4string <- all_crs[["CRS"]]
  return(x)
})


# ==== CRS ====

#' @export
#' @rdname projection
setMethod("crs", "LASheader", function(x, asText = FALSE)
{
  return(projection(x, asText))
})

#' @export
#' @rdname projection
setMethod("crs", "LAS", function(x, asText = FALSE)
{
  return(projection(x, asText))
})

#' @export
#' @rdname projection
setMethod("crs<-", "LAS", function(x, ..., value)
{
  projection(x) <- value
  return(x)
})


#' @export
#' @rdname projection
setMethod("crs", "LAScatalog", function(x, asText = FALSE)
{
  return(projection(x, asText))
})

#' @export
#' @rdname projection
setMethod("crs<-", "LAScatalog", function(x, ..., value)
{
  projection(x) <- value
  return(x)
})


# ===== SPTRANSFORM ====

#' @importMethodsFrom sp spTransform
#' @noRd
setMethod("spTransform", signature("LAS", "CRS"), function(x, CRSobj, ...)
{
  if (is.na(x@proj4string@projargs))
    stop("No transformation possible from NA reference system")

  p <- list(...)

  # Transform the point coordinates
  spts <- sp::SpatialPoints(coordinates(x))
  spts@proj4string <- crs(x)
  spts <- sp::spTransform(spts, CRSobj)
  X <- spts@coords[,1]
  Y <- spts@coords[,2]

  # Update the offsets in the header
  offsetx <- floor(min(X))
  offsety <- floor(min(Y))
  if (!is.null(p$xoffset)) offsetx <- p$xoffset
  if (!is.null(p$yoffset)) offsety <- p$yoffset
  x@header@PHB[["X offset"]] <- offsetx
  x@header@PHB[["Y offset"]] <- offsety

  # Update the scale factors
  scalex <- x@header@PHB[["X scale factor"]]
  scaley <- x@header@PHB[["Y scale factor"]]
  if (!is.null(p$scale)) scalex <- scaley <- p$scale
  x@header@PHB[["X scale factor"]] <- scalex
  x@header@PHB[["Y scale factor"]] <- scaley

  # Quantize the coordinates
  fast_quantization(X, scalex, offsetx)
  fast_quantization(Y, scaley, offsety)

  # Update the LAS object
  x@data[["X"]] <- X
  x@data[["Y"]] <- Y
  x <- lasupdateheader(x)
  crs(x) <- CRSobj

  return(x)
})

# ===== INTERNAL TOOLS =======

use_wktcs <- function(x) {
  UseMethod("use_wktcs", x)
}

use_wktcs.LAS <- function(x) {
  return(x@header@PHB[["Global Encoding"]][["WKT"]])
}

use_wktcs.LASheader <- function(x) {
  return(x@PHB[["Global Encoding"]][["WKT"]])
}

use_epsg <- function(x) {
  UseMethod("use_epsg", x)
}

use_epsg.LAS <- function(x) {
  return(!x@header@PHB[["Global Encoding"]][["WKT"]])
}

use_epsg.LASheader <- function(x) {
  return(!x@PHB[["Global Encoding"]][["WKT"]])
}

all_crs_formats = function(value)
{
  if (is(value, "CRS"))
  {
    proj4 <- value@projargs
    CRS <- value
    crs <- sf::st_crs(CRS)
    wkt <- crs$wkt
    epsg <- crs$epsg
  }
  else if (is(value, "crs"))
  {
    crs <- value
    wkt <- value$wkt
    CRS <- as(crs, "CRS")
    proj4 <- CRS@projargs
    epsg <- crs$epsg
  }
  else if (is.character(value))
  {
    crs <- sf::st_crs(value)
    CRS <- as(crs, "CRS")
    proj4 <- CRS@projargs
    wkt <- crs$wkt
    epsg <- crs$epsg
  }
  else if (is.numeric(value))
  {
    epsg <- value
    crs <- sf::st_crs(epsg)
    CRS <- as(crs, "CRS")
    proj4 <- CRS@projargs
    wkt <- crs$wkt
  }
  else
    stop("'value' is not a CRS or a string or a number.")

  return(list(epsg = epsg, CRS = CRS, crs = crs, proj4 = proj4, wkt = wkt))
}

epsg2CRS <- function(epsg, fail = FALSE)
{
  crs <- tryCatch(
  {
    crs <- suppressWarnings(sf::st_crs(paste0("EPSG:", epsg)))
    return(suppressWarnings(as(crs, "CRS")))
  },
  error = function(e)
  {
    if (!fail)
      return(sp::CRS(NA_character_))
    else
      stop(paste("Invalid epsg code", epsg), call. = FALSE)
  })

  return(crs)
}

wkt2CRS <- function(wkt, fail = FALSE)
{
  crs <- tryCatch(
  {
    as(sf::st_crs(wkt), "CRS")
  },
  error = function(e)
  {
    if (!fail)
      return(sp::CRS(NA_character_))
    else
      stop("Invalid WKT string", call. = FALSE)
  })

  return(crs)
}
